"""Module for reading/writing ini style files"""
# vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4 foldmethod=indent:
import os
import configparser
import time
from shutil import copyfile
import errno
import logging
import threading

log = logging.getLogger(__name__)
# log.setLevel(logging.DEBUG)


class IniFile(configparser.ConfigParser):
    """Overidden ConfigParser class to read and write ini format files"""

    def __init__(self, filename, takebackup=True, interpolation=None):
        """Reads filename and parses it into a configparser config structure

        Arguments:
            filename - name of the ini format file to read and parse
        """
        # call the parent constructor
        super().__init__(interpolation=interpolation)
        self.dirty = False
        self.filename = filename
        self.takebackup = takebackup
        self.dir = os.path.dirname(self.filename)
        self.backupdir = "{}/backup".format(self.dir)
        self.backupfn = False
        log.debug(
            "IniFile Class: dir: {}, backup dir: {}".format(self.dir, self.backupdir)
        )
        if not os.path.exists(filename):
            log.error("INI file: {} does not exist".format(self.filename))
            raise FileExistsError("INI file: {} does not exist".format(self.filename))
        self.makeBackup()
        # fp=open(self.filename,"r")
        self.read(self.filename)
        self._lock = threading.Lock()
        # fp.close()

    def makeBackup(self):
        if self.takebackup and self.makeDir(self.backupdir):
            self.backupfn = "{}/{}-{}".format(
                self.backupdir,
                os.path.basename(self.filename),
                time.strftime("%Y%m%d%H%M%S"),
            )
            log.debug("backup fn: {}".format(self.backupfn))
            link = "{}/previous-{}".format(self.dir, os.path.basename(self.filename))
            log.debug("previous symlink: {}".format(link))
            copyfile(self.filename, self.backupfn)
            try:
                os.symlink(self.backupfn, link)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    os.remove(link)
                    os.symlink(self.backupfn, link)
                else:
                    raise

    def backupFN(self):
        return self.backupfn

    def makeDir(self, newdir):
        try:
            os.makedirs(newdir)
        except OSError as e:
            # if it exists do nothing
            if e.errno != errno.EEXIST:
                # permissions, disk full etc
                raise
        return True

    def titles(self):
        sectionlist = self.sections()
        return sectionlist

    def sectionExists(self, section):
        ret = False
        titles = self.titles()
        if section in titles:
            ret = True
        return ret

    def getSectionItems(self, section):
        # section items look like
        # [('region', 'eu-west-1'), ('aws_access_key_id', 'xxxxx'), ('aws_secret_access_key', 'xxxxxx')]
        # so, we need to create another dictionary based on the keys
        items = self.items(section)
        k = {}
        for i in range(0, len(items)):
            k[items[i][0]] = items[i][1]
        log.debug("section: {}, items: {}".format(section, items))
        return k

    def updateSection(self, section, datadict, write=False):
        ret = False
        try:
            for k in datadict:
                self.set(section, k, datadict[k])
                log.debug(
                    "Setting {} to {} for section {}".format(k, datadict[k], section)
                )
            if write:
                self.saveData()
            ret = True
        except Exception as e:
            log.error(
                "Exception occurred when updating section: {}, {}".format(section, e)
            )
            raise
        return ret

    def deleteSection(self, section):
        self.remove_section(section)
        self.saveData()

    def saveData(self):
        try:
            with self._lock:
                log.debug("writing out ini file")
                fp = open(self.filename, "w")
                self.write(fp)
                fp.close()
        except Exception as e:
            msg = f"Exception in inifile:saveData: {e}"
            print(msg)
            raise



import hashlib
import os
import random

# initialise the random number generator
random.seed()


def pwgen(length=15, nospaces=False):
    # acceptable list of chars that can be capitalised
    capl = "acdefghjklmnpqrstuvwxyz"
    # list of acceptable symbols
    symbols = "=-!$%^&*(){}[]"
    # find the current directory name
    wlfn = "/".join([os.path.dirname(__file__), "wordlist.txt"])
    # read the words into a list (note: each word will still have the
    # trailing newline)
    with open(wlfn, "r") as wl:
        wlist = wl.readlines()
    # pick 4 random words from the list
    words = [word.rstrip() for word in random.choices(wlist, k=4)]
    # join the words into a phrase
    phrase = " ".join(words)
    # convert to bytes and hash into 64 chars.
    hphrase = hashlib.sha256(phrase.encode()).hexdigest()
    # pick 15 random characters from the hashed phrase
    ch = "".join(random.choices(hphrase, k=length))
    # pick a letter in the string to capitalise
    # filter out letters that are not in our acceptable list
    pos = [ch.index(c) for c in ch if c in capl]
    pi = random.choice(pos)
    # split the string apart at our chosen letter
    # capitalise the letter and re-combine the string
    left = ch[: pi - 1] if pi > 0 else ""
    c = ch[pi].upper()
    right = ch[pi + 1 :] if pi < len(ch) - 1 else ""
    pw = f"{left}{c}{right}"
    # pick a random position within that password
    # where we can insert a symbol
    pi = random.randint(0, len(pw) - 1)
    # string slicing again
    left = pw[:pi] if pi > 0 else ""
    right = pw[pi:] if pi < len(pw) - 1 else ""
    # pick a random symbol
    symbol = random.choice(symbols)
    # create the final password
    pw = f"{left}{symbol}{right}"
    # split the password into blocks of 4 chars.
    xpw = [pw[i : i + 4] for i in range(0, len(pw), 4)]
    # join them together into one string seperated by a space
    if nospaces:
        op = "".join(xpw)
    else:
        op = " ".join(xpw)
    # display our work to the world
    return op


if __name__ == "__main__":
    pwgen()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ccaawscreds']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.14.48,<2.0.0', 'click>=7.1.2,<8.0.0']

entry_points = \
{'console_scripts': ['awskey = ccaawscreds.credentials:awsK',
                     'awspw = ccaawscreds.credentials:awsPW',
                     'genpw = ccaawscreds.credentials:genPW']}

setup_kwargs = {
    'name': 'ccaawscreds',
    'version': '0.5.2',
    'description': 'update AWS console PW, rotate Access Keys, generate random passwords.',
    'long_description': None,
    'author': 'ccdale',
    'author_email': 'chris.charles.allison@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""IAM user module for the awscreds application."""

from datetime import datetime

import boto3
from botocore.exceptions import ProfileNotFound


class AccessKeyError(Exception):
    pass


class IAMUser:
    def __init__(self, username, profile):
        try:
            self.profileok = False
            if profile is None:
                self.client = boto3.client("iam")
                self.profileok = True
            else:
                sess = boto3.session.Session(profile_name=profile)
                self.client = sess.client("iam")
                self.profileok = True
            self.founduser = self.getUserInfo(username)
            self.newkey = None
        except ProfileNotFound as e:
            print(f"aws account profile: {profile} does not exist")
        except Exception as e:
            msg = f"Exception in IAMUser.__init__: {type(e).__name__}: {e}"
            print(msg)
            raise

    def getUserInfo(self, username):
        try:
            self.userinfo = self.client.get_user(UserName=username)["User"]
            self.userinfo["passwordage"] = self.getPasswordAge()
            return True
        except self.client.exceptions.NoSuchEntityException as e:
            print(f"user: {username} does not exist.")
            return False
        except Exception as e:
            msg = f"Exception in IAMUser.getUserInfo: {type(e).__name__}: {e}"
            print(msg)
            raise

    def getPasswordAge(self):
        try:
            passwordage = None
            lp = self.client.get_login_profile(UserName=self.userinfo["UserName"])
            if lp is not None:
                if "LoginProfile" in lp:
                    passwordage = lp["LoginProfile"]["CreateDate"]
            return passwordage
        except self.client.exceptions.NoSuchEntityException as e:
            return None
        except Exception as e:
            msg = f"Exception in IAMUser.getPasswordAge: {type(e).__name__}: {e}"
            print(msg)
            raise

    def changePassword(self, oldpw, newpw):
        """Changes the calling users password.

        Relies on boto3 raising an exception if it fails
        """
        try:
            if self.profileok and self.founduser:
                kwargs = {"OldPassword": oldpw, "NewPassword": newpw}
                pwchanged = self.client.change_password(**kwargs)
                return True
        except Exception as e:
            msg = f"Exception in IAMUser.changePassword: {type(e).__name__}: {e}"
            print(msg)
        return False

    def getKeys(self, username):
        """returns a user dict complete with registered access keys."""
        try:
            keys = self.client.list_access_keys(UserName=username)
            okeys = []
            for key in keys["AccessKeyMetadata"]:
                kid = key["AccessKeyId"]
                klu = self.client.get_access_key_last_used(AccessKeyId=kid)
                if "LastUsedDate" in klu["AccessKeyLastUsed"]:
                    key["LastUsedDate"] = klu["AccessKeyLastUsed"]["LastUsedDate"]
                    key["lastused"] = int(
                        datetime.timestamp(klu["AccessKeyLastUsed"]["LastUsedDate"])
                    )
                okeys.append(key)
            return okeys
        except Exception as e:
            msg = f"Exception in IAMUser.getKeys: {e}"
            msg += f"Failed to obtain access key information for user: {username}"
            raise Exception(msg)

    def __delete_inactive_keys(self, username):
        """If the user has an inactive key, this will delete it."""
        try:
            activekeys = 0
            keys = self.getKeys(username)
            for key in keys:
                if key["Status"] == "Inactive":
                    self.client.delete_access_key(
                        AccessKeyId=key["AccessKeyId"], UserName=username
                    )
                else:
                    activekeys += 1
            if activekeys > 1:
                raise AccessKeyError(
                    f"user {username} has more than one active access key, please delete one and try again"
                )
        except Exception as e:
            msg = f"Exception in IAMUser.__delete_inactive_keys: {e}"
            msg += f"""\nAn error occurred deleting the inactive access key: {key["AccessKeyId"]}"""
            msg += f"\nfor user {username}"
            print(msg)
            raise

    def __deactivate_key(self, username, keyid):
        """Deactivates the keyid key."""
        try:
            self.client.update_access_key(
                AccessKeyId=keyid, UserName=username, Status="Inactive"
            )
            return True
        except Exception as e:
            msg = f"Exception in IAMUser.__deactivate_key: {e}"
            msg += f"Failed to deactivate the key {keyid} for user {username}"
            print(msg)
        return False

    def __generate_new_key(self, username):
        key = False
        try:
            self.newkey = self.client.create_access_key(UserName=username)
            return True
        except Exception as e:
            msg = f"Exception in IAMUser.__generate_new_key: {e}"
            print(msg)
            raise
        return False

    def rotateKeys(self, username, ckeyid):
        """This will leave the user with a new, active key

        It first deletes any inactive key,
        then generates a new active key,
        the de-activates the original key.
        returns a dictionary of the new key:
        {'AccessKey':
            {'UserName': 'chris.allison',
            'AccessKeyId': 'AKIAXIPAILAXFADFIWPV',
            'Status': 'Active',
            'SecretAccessKey': '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',
            'CreateDate': datetime.datetime(2020, 9, 2, 16, 32, 4, tzinfo=tzutc())},
            'ResponseMetadata': {...}
        }
        """
        try:
            self.__delete_inactive_keys(username)
            # at this point we should only have one key, the current, active one
            # passed in as ckeyid param.
            self.__generate_new_key(username)
            keys = self.getKeys(username)
            if len(keys) > 0:
                for key in keys:
                    if key["AccessKeyId"] == ckeyid:
                        self.__deactivate_key(username, ckeyid)
            return self.newkey
        except Exception as e:
            msg = f"Exception in IAMUser.rotateKeys: {e}"
            print(msg)
            raise



"""awscreds Credentials module."""

import os

import boto3
import click

from ccaawscreds import __version__
from ccaawscreds.iamuser import IAMUser
from ccaawscreds.inifile import IniFile
from ccaawscreds.pwgen import pwgen


@click.group()
def awsk():
    pass


@click.group()
def awspw():
    pass


@click.group()
def genpw():
    pass


def findConfigFile():
    try:
        ini = None
        fn = os.path.expanduser("~/.aws/config")
        if os.path.exists(fn):
            ini = IniFile(fn)
        return ini
    except Exception as e:
        msg = f"Exception in findConfigFile: {type(e).__name__}: {e}"
        print(msg)
        raise


def setup(profile):
    """Returns the iam and ini objects, finds the named profile and user.

    will raise exceptions if any of that fails
    """
    try:
        ini = user = iam = xprofile = None
        print(f"awscreds {__version__}")
        ini = findConfigFile()
        if ini is not None:
            xprofile = f"profile {profile}"
            if ini.sectionExists(xprofile):
                items = ini.getSectionItems(xprofile)
                user = items["user"]
                iam = IAMUser(user, profile)
                if not iam.profileok:
                    msg = f"Failed to locate profile: {profile} in config file"
                    raise Exception(msg)
                if not iam.founduser:
                    msg = f"cannot find the 'user' key in profile {profile}"
                    raise Exception(msg)
            else:
                msg = f"profile {profile} not found in config file ~/.aws/config"
                raise Exception(msg)
        else:
            msg = "Config file not found: ~/.aws/config"
            raise Exception(msg)
        return (ini, iam, user, xprofile)
    except Exception as e:
        msg = f"Exception in setup: {type(e).__name__}: {e}"
        print(msg)
        raise


@awsk.command()
@click.argument("profile")
def awsK(profile):
    """Rotate the users access key for the named profile in the ~/.aws/config file."""
    try:
        ini, iam, user, xprofile = setup(profile)
        if (
            ini is not None
            and iam is not None
            and user is not None
            and xprofile is not None
        ):
            items = ini.getSectionItems(xprofile)
            akid = items["aws_access_key_id"]
            newkey = iam.rotateKeys(user, akid)
            # print(newkey)
            ak = newkey["AccessKey"]
            items["aws_access_key_id"] = ak["AccessKeyId"]
            items["aws_secret_access_key"] = ak["SecretAccessKey"]
            ini.updateSection(xprofile, items, True)
            print(f"""New access key id {items["aws_access_key_id"]} created.""")
    except Exception as e:
        msg = f"Exception in awsK: {type(e).__name__}: {e}"
        print(msg)


@awspw.command()
@click.option(
    "--length",
    "-l",
    default="16",
    help="overall length of the new password (default: 16)",
)
@click.argument("profile")
def awsPW(profile, length):
    """Change the users AWS Console Password for the named profile in the ~/.aws/config file."""
    try:
        ilength = int(length)
        if ilength > 32:
            ilength = 32
        if ilength < 8:
            ilength = 8
        ini, iam, user, xprofile = setup(profile)
        if (
            ini is not None
            and iam is not None
            and user is not None
            and xprofile is not None
        ):
            items = ini.getSectionItems(xprofile)
            newpw = pwgen(ilength)
            iam.changePassword(items["password"], newpw)
            print(f"""current: {items["password"]}""")
            items["password"] = newpw
            ini.updateSection(xprofile, items, True)
            print(f"new: {newpw}")
    except Exception as e:
        msg = f"Exception in awsPW: {type(e).__name__}: {e}"
        print(msg)


@genpw.command()
@click.option(
    "--length",
    "-l",
    default="16",
    help="overall length of the new password (default: 16)",
)
@click.option(
    "--nospaces",
    "-n",
    default=False,
    is_flag=True,
    help="do not put spaces in the generated password. Default is to chunk the new password into blocks of 4 characters, seperated by spaces.",
)
def genPW(length, nospaces):
    """Generate a new password of `length` characters. (limited to between 8 and 32 - default is 16)."""
    try:
        ilength = int(length)
        if ilength > 32:
            ilength = 32
        if ilength < 8:
            ilength = 8
        newpw = pwgen(ilength)
        if nospaces:
            print(newpw.replace(" ", ""))
        else:
            print(newpw)
    except Exception as e:
        msg = f"Exception in genPW: {type(e).__name__}: {e}"
        print(msg)
